/**
 * Copyright (c) Perforce Software, Inc., 1997-2009. All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE
 * SOFTWARE, INC. BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */
package com.perforce.ant.tasks;

import com.perforce.p4java.client.P4JClient;
import com.perforce.p4java.core.P4JChangeList;
import com.perforce.p4java.core.file.P4JFileSpec;
import com.perforce.p4java.core.file.P4JFileSpecBuilder;
import com.perforce.p4java.core.file.P4JFileSpecOpStatus;
import com.perforce.p4java.exception.P4JException;
import com.perforce.p4java.impl.generic.core.P4JChangeListImpl;
import com.perforce.p4java.impl.generic.core.file.P4JFileSpecImpl;
import com.perforce.p4java.server.P4JServer;
import com.perforce.p4java.server.P4JServerFactory;
import com.perforce.p4java.server.callback.P4JCommandCallback;

import java.io.File;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Copy;
import org.apache.tools.ant.taskdefs.Delete;
import org.apache.tools.ant.types.FileSet;

/**
 * Task to replace the content of a folder in the depot with a specifed folder.
 * This task will add new files, delete deleted files, and edit modified files.
 * 
 * This task submits the changes with the configured comment.
 * 
 * @author Kevin Sawicki (ksawicki@perforce.com)
 * 
 */
public class PerforceReplaceTask extends Task {

    private String fromDir = null;
    private String toDir = null;
    private String port = null;
    private String client = null;
    private String user = null;
    private String comment = null;

    /**
     * Set the comment that will be used as the changelist description. This is
     * an optional field and if unset a default will be used.
     * 
     * @param comment
     *            the comment to set
     */
    public void setComment(String comment) {
	this.comment = comment;
    }

    /**
     * Set the directory to copy the files from.
     * 
     * @param fromDir
     *            the fromDir to set
     */
    public void setFromDir(String fromDir) {
	this.fromDir = fromDir;
    }

    /**
     * 
     * Set the directory to copy the files to.
     * 
     * @param toDir
     *            the toDir to set
     */
    public void setToDir(String toDir) {
	this.toDir = toDir;
    }

    /**
     * Set the Perforce server address and port to use.
     * 
     * @param port
     *            - Perforce server address, i.e localhost:1666
     */
    public void setPort(String port) {
	this.port = port;
    }

    /**
     * Set the Perforce client workspace to use.
     * 
     * @param client
     *            - Perforce client workspace name
     */
    public void setClient(String client) {
	this.client = client;
    }

    /**
     * Set the Perforce user to use.
     * 
     * @param user
     *            - Perforce user name
     */
    public void setUser(String user) {
	this.user = user;
    }

    private void loadP4Variables() {
	if (port == null) {
	    port = getProject().getProperty("p4.port.replace");
	    if (port == null) {
		port = System.getenv("P4PORT");
	    }
	}
	if (client == null) {
	    client = getProject().getProperty("p4.client.replace");
	    if (client == null) {
		client = System.getenv("P4CLIENT");
	    }
	}
	if (user == null) {
	    user = getProject().getProperty("p4.user.replace");
	    if (user == null) {
		user = System.getenv("P4USER");
	    }
	}
	if (comment == null) {
	    comment = "Replacing files at " + toDir;
	}
    }

    /**
     * @see org.apache.tools.ant.Task#execute()
     */
    public void execute() throws BuildException {
	try {
	    loadP4Variables();
	    if (fromDir == null) {
		throw new BuildException("fromDir must be set");
	    }
	    if (toDir == null) {
		throw new BuildException("toDir must be set");
	    }
	    if (port == null) {
		throw new BuildException(
			"port must be set or the p4.port.replace property set or the P4PORT environment variable set");
	    }
	    if (user == null) {
		throw new BuildException(
			"user must be set or the p4.user.replace property set or the P4USER environment variable set");
	    }
	    if (client == null) {
		throw new BuildException(
			"client must be set or the p4.client.replace property set or the P4CLIENT environment variable set");
	    }
	    replace();
	} catch (Throwable t) {
	    t.printStackTrace();
	}
    }

    private void replace() throws BuildException {
	File toDirectory = new File(toDir);
	if (toDirectory.exists()) {
	    Delete delete = new Delete();
	    delete.setProject(getProject());
	    delete.setIncludeEmptyDirs(true);
	    FileSet deleteSet = new FileSet();
	    deleteSet.setDir(toDirectory);
	    deleteSet.setIncludes("**/*");
	    delete.addFileset(deleteSet);
	    delete.execute();
	} else {
	    toDirectory.mkdirs();
	}

	Copy copy = new Copy();
	copy.setProject(getProject());
	copy.setTodir(new File(toDir));
	FileSet copySet = new FileSet();
	copySet.setDir(new File(fromDir));
	copySet.setIncludes("**/*");
	copy.addFileset(copySet);
	copy.execute();

	try {
	    runP4Commands();
	} catch (P4JException e) {
	    throw new BuildException("P4JException thrown", e);
	}
    }

    private String addEllipsis(String path) {
	StringBuilder builder = new StringBuilder(path);
	if (path.length() > 0) {
	    if (path.charAt(path.length() - 1) != (File.separatorChar)) {
		builder.append(File.separatorChar);
	    }
	}
	builder.append("...");
	return builder.toString();
    }

    private P4JClient createClient() throws P4JException {
	P4JClient client = null;
	try {
	    P4JServer server = P4JServerFactory.getServer("p4java://" + port,
		    null);
	    server.registerCallback(new P4JCommandCallback() {

		public void receivedServerMessage(int key, int genericCode,
			int severityCode, String message) {
		}

		public void receivedServerInfoLine(int key, String infoLine) {
		    getProject().log(infoLine, Project.MSG_INFO);
		}

		public void receivedServerErrorLine(int key, String errorLine) {
		    getProject().log(errorLine, Project.MSG_ERR);
		}

		public void issuingServerCommand(int key, String commandString) {
		    getProject().log(commandString);
		}

		public void completedServerCommand(int key, long millisecsTaken) {

		}
	    });
	    server.connect();
	    server.setUserName(user);
	    client = server.getClient(this.client);
	    server.setCurrentClient(client);
	} catch (URISyntaxException e) {
	    throw new P4JException(e);
	}
	if (client == null) {
	    throw new P4JException("Null client returned from P4JServer");
	}
	return client;
    }

    private List<P4JFileSpec> getDifferingFiles(P4JClient client,
	    List<P4JFileSpec> specs) throws P4JException {
	List<P4JFileSpec> editList = new ArrayList<P4JFileSpec>();
	List<P4JFileSpec> diffList = client.getDiffFileList(specs, 0, false,
		false, false, false, true, false, false);
	for (P4JFileSpec spec : diffList) {
	    if (spec.getOpStatus() == P4JFileSpecOpStatus.VALID) {
		editList.add(new P4JFileSpecImpl(spec.getPreferredPath(false)));
	    }
	}
	return editList;
    }

    private List<P4JFileSpec> getMissingFiles(P4JClient client,
	    List<P4JFileSpec> specs) throws P4JException {
	List<P4JFileSpec> missing = client.getDiffFileList(specs, 0, false,
		false, false, true, false, false, false);
	List<P4JFileSpec> deleteList = new ArrayList<P4JFileSpec>();
	for (P4JFileSpec spec : missing) {
	    if (spec.getOpStatus() == P4JFileSpecOpStatus.VALID) {
		deleteList
			.add(new P4JFileSpecImpl(spec.getPreferredPath(false)));
	    }
	}
	return deleteList;
    }

    private P4JChangeList createChangelist(P4JClient client)
	    throws P4JException {
	P4JChangeListImpl cl = new P4JChangeListImpl();
	cl.setId(P4JChangeList.UNKNOWN);
	cl.setClientId(this.client);
	cl.setDescription(this.comment);
	cl.setUsername(this.user);

	P4JChangeList created = client.newChangeList(cl);
	if (created == null) {
	    throw new P4JException(
		    "Null new changelist returned from P4JClient");
	}
	return created;
    }

    private void addFiles(List<P4JFileSpec> adds, File directory) {
	for (File file : directory.listFiles()) {
	    if (file.isDirectory()) {
		addFiles(adds, file);
	    } else {
		adds.add(new P4JFileSpecImpl(file.getAbsolutePath()));
	    }
	}
    }

    private List<P4JFileSpec> getNewFiles() {
	List<P4JFileSpec> adds = new ArrayList<P4JFileSpec>();
	File directory = new File(toDir);
	if (directory.isDirectory()) {
	    addFiles(adds, directory);
	}
	return adds;
    }

    private void runP4Commands() throws P4JException {
	String toSpec = addEllipsis(toDir);
	List<P4JFileSpec> specs = P4JFileSpecBuilder
		.makeFileSpecList(new String[] { toSpec });
	P4JClient client = createClient();
	P4JChangeList list = createChangelist(client);
	try {
	    List<P4JFileSpec> addList = getNewFiles();
	    client.addFiles(addList, false, list.getId(), null, true);

	    List<P4JFileSpec> editList = getDifferingFiles(client, specs);
	    if (editList.size() > 0) {
		client.editFiles(editList, false, list.getId(), null);
	    }

	    List<P4JFileSpec> deleteList = getMissingFiles(client, specs);
	    if (deleteList.size() > 0) {
		client.deleteFiles(deleteList, list.getId(), false);
	    }

	    client.revertFiles(specs, false, list.getId(), true, false);

	    list.refresh();
	    list.submit(false);
	} finally {
	    // Revert files and delete changelist if any exceptions thrown
	    if (list != null) {
		try {
		    client.revertFiles(P4JFileSpecBuilder
			    .makeFileSpecList(new String[] { "//"
				    + client.getName() + "/..." }), false, list
			    .getId(), false, false);
		    client.getServer().deletePendingChangeList(list.getId());
		} catch (P4JException e2) {
		    // Suppress the second one and let the first be re-thrown
		}
	    }
	}

    }
}
